#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/fcntl.h>
#include <sys/stat.h>

#include "common.h"

struct file1
{
    char        type;   // 0 - unknown, 1 - file, 2 - directory
    char        hs;     // "has subdirs" flag (directories only)
    time_t      age;    // st_ctime (useful for sorting)
    long        size;   // in bytes
    int         dir_offset;   // directory where thing is
    int         name_offset;  // file/directory name
    int         desc_offset;  // description (if available)
};

/*--------------------------------------------------------------------*/

static int   fh_main, fh_name, fh_desc;
static int   name_off=0, desc_off=0;
static char  *names, *descriptions;
static int   dir_off, total_counter = 0;

struct __fb fb;

#ifdef __EMX__
#define strcasecmp stricmp
#define lstat stat
#endif

static int attach_descriptions (struct file *f, int n);
static int process_directory (char *directory);
static int filecompare1 (__const__ void *elem1, __const__ void *elem2);

/*--------------------------------------------------------------------*/

int load_database (void)
{
    struct file1   *files1;
    int  i;
    long L;

    // check for lockfile
    
    // load files into memory
    if (load_file (NAMELIST, &names) < 0) return 1;
    if (load_file (DESCLIST, &descriptions) < 0) return 1;
    if ((L=load_file (FILELIST, (char **) &files1)) < 0) return 1;

    // make 'files' from 'files1'
    fb.N = L / sizeof (struct file1);
    fb.files = malloc (fb.N * sizeof (struct file));
    for (i=0; i<fb.N; i++)
    {
        fb.files[i].type = files1[i].type;
        fb.files[i].hs = files1[i].hs;
        fb.files[i].age = files1[i].age;
        fb.files[i].size = files1[i].size;
        fb.files[i].dir  = names + files1[i].dir_offset;
        fb.files[i].name = names + files1[i].name_offset;
        if (files1[i].desc_offset != 0)
            fb.files[i].desc = descriptions + files1[i].desc_offset;
        else
            fb.files[i].desc = NULL;
    }
    free (files1);
    fb.timestamp = * ((time_t *) (descriptions + 2));
    return 0;
}

/*--------------------------------------------------------------------*/

#define DEF_PERM S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH

int create_database (char *sd)
{
    time_t  now;
    char    filelist[4096], namelist[4096], desclist[4096];

    sprintf (filelist, "%s.1", FILELIST);
    sprintf (namelist, "%s.1", NAMELIST);
    sprintf (desclist, "%s.1", DESCLIST);
    
    fh_main = open (filelist, O_CREAT|O_TRUNC|O_RDWR, DEF_PERM);
    fh_name = open (namelist, O_CREAT|O_TRUNC|O_RDWR, DEF_PERM);
    fh_desc = open (desclist, O_CREAT|O_TRUNC|O_RDWR, DEF_PERM);
    if (fh_main < 0 || fh_name < 0 || fh_desc < 0) return 1;

    // empty description and timestamp
    write (fh_desc, "$", 2);
    desc_off += 2;
    now = time (NULL);
    write (fh_desc, &now, sizeof (time_t));
    desc_off += sizeof (time_t);

    // walk directory tree
    process_directory (sd);

    close (fh_main);
    close (fh_name);
    close (fh_desc);
    fprintf (cgiOut, "\nEntries processed: %d", total_counter);

    // create lock file
    close (open (LOCKFILE, O_RDWR|O_CREAT|O_EXCL, 00700));

    rename (filelist, FILELIST);
    rename (namelist, NAMELIST);
    rename (desclist, DESCLIST);
    
    // remove lock file
    remove (LOCKFILE);
    
    return 0;
}

/*--------------------------------------------------------------------*/

int write_name (char *txt)
{
    int  l = strlen (txt)+1;
    
    if (write (fh_name, txt, l) != l) return -1;
    name_off += l;
    return name_off-l;
}

/*--------------------------------------------------------------------*/

int write_desc (char *txt)
{
    int  l = strlen (txt)+1;
    
    if (write (fh_desc, txt, l) != l) return -1;
    desc_off += l;
    return desc_off-l;
}

/*--------------------------------------------------------------------*/

int write_item (struct file1 *f1)
{
    total_counter++;
    if (write (fh_main, f1,
               sizeof (struct file1)) != sizeof (struct file1)) return -1;
    return 0;
}

/*--------------------------------------------------------------------*/

static int process_directory (char *directory)
{
    int               i, N, N_d, N_f, N_l;
    char              buf[2048];
    struct file       *f;
    struct file1      f1;
    
    // add entry to DIRLIST file
    dir_off = write_name (directory);

    // build list of files
    f = (struct file *) malloc (MAX_ENTRIES * sizeof (struct file));
    N = MAX_ENTRIES;
    get_directory (directory, f, &N);
    
    // separate directories from files
    N_d = N_l = N_f = 0;
    for (i=0; i<N; i++)
    {
        if (f[i].type == TYPE_DIRECTORY) N_d++;
        if (f[i].type == TYPE_SYMLINK) N_l++;
        if (f[i].type == TYPE_REGULAR) N_f++;
    }

    fprintf (cgiOut, "%s : %d/%d/%d dirs/links/files\n", directory, N_d, N_l, N_f);

    // write items
    for (i=0; i<N; i++)
    {
        f1.type = f[i].type;
        f1.hs   = f[i].hs;
        f1.age  = f[i].age;
        f1.size = f[i].size;
        f1.dir_offset  = dir_off;
        f1.name_offset = write_name (f[i].name);
        f1.desc_offset = f[i].desc == NULL ? 0 : write_desc (f[i].desc);
        if (write_item (&f1)) return 1;
    }

    // process directories (do not follow symlinks!)
    for (i=0; i<N; i++)
    {
        if (f[i].type != TYPE_DIRECTORY) continue;
        strcpy (buf, directory);
        strcats (buf, f[i].name);
        process_directory (buf);
    }
    free (f);
    return 0;
}

/*--------------------------------------------------------------------*/

static int attach_descriptions (struct file *f, int n)
{
    int   i, j, nd, type;
    FILE *fp;
    char buf1[4096], buf2[4096], buf3[8192];
    char *pname, *pdesc, *p, *p1, *dirs[1024];

    // build list of directories to look at
    nd = 0;
    dirs[nd] = f[0].dir;
    for (i=0; i<n; i++)
    {
        for (j=0; j<nd; j++)
            if (strcmp (f[i].dir, dirs[j]) == 0) break;
        if (j == nd)
            dirs[nd++] = f[i].dir;
    }

    // load all index files and attach descriptions where approriate
    for (j=0; j<nd; j++)
    {
        // try to find index file
        strcpy (buf1, make_absolute (dirs[j]));

        type = 0; // default. description follows filename (LEO and SAI)
        strcpy (buf2, buf1);
        strcats (buf2, "00-index");
        if (access (buf2, 00) == 0) goto found;

        type = 1; // Format: filename date description (HOBBES)
        strcpy (buf2, buf1);
        strcats (buf2, "00index.txt");
        if (access (buf2, 00) == 0) goto found;

        // no index file found
        break;

    found:
        // read file line-by-line, trying to find where descriptions belong
        fp = fopen (buf2, "rt");
        if (fp == NULL) break;
    
        while (fgets (buf3, sizeof(buf3), fp) != NULL)
        {
            p = buf3;
            // remove CR/LF at the end
            strip_nl (buf3);
            // skip whitespace at the beginning
            while (*p == ' ' && *p != '\0') p++;
            if (*p == '\0') continue;
            pname = p;
            // find where name ends
            while (*p != ' ' && *p != '\0') p++;
            if (*p == '\0') continue;
            if (*(p-1) == '/')
                *(p-1) = '\0';    // directories have ending '/'
            else
                *p = '\0';
            p++;
            switch (type)
            {
            case 0: // LEO and SAI
                p1 = p;
                while (*p1 != '\0' && isspace(*p1)) p1++; // skip whitespace
                if (isdigit(*p1) && isdigit (*(p1+1))) // skip size if present
                    while (isdigit(*p1) && *p1 != '\0') p1++;
                break;
            case 1: // HOBBES
                p1 = strsep1 (&p, ' '); // p1 now points to file size
                if (p1 == NULL) continue;
                p1 = strsep1 (&p, ' '); // p1 now points to month
                if (p1 == NULL) continue;
                p1 = strsep1 (&p, ' '); // p1 now points to day
                if (p1 == NULL) continue;
                p1 = strsep1 (&p, ' '); // p1 now points to year/time
                if (p1 == NULL) continue;
                p1 = p;
                while (*p1 != '\0' && isspace(*p1)) p1++; // skip whitespace
                break;
            default:
                p1 = ""; // to keep compiler happy
            }
            pdesc = p1;
            // when description is empty, skip it
            if (strlen (pdesc) == 0) continue;
            // find to what file it belongs
            for (i=0; i<n; i++)
                if ((f[i].type == TYPE_REGULAR ||
                     f[i].type == TYPE_DIRECTORY) &&
                    !strcmp (f[i].name, pname) &&
                    !strcmp (f[i].dir, dirs[j]))
                {
                    f[i].desc = chunk_put (pdesc);
                    break;
                }
        }
        fclose (fp);
    }

    // directory descriptions are to be handled separately
    for (i=0; i<n; i++)
        if (f[i].type == TYPE_DIRECTORY && f[i].desc == NULL)
        {
            strcpy (buf1, f[i].dir);
            strcats (buf1, f[i].name);
            p = dir_description (buf1);
            if (p != NULL)
                f[i].desc = chunk_put (p);
        }
    
    return 0;
}

/*--------------------------------------------------------------------*/

int get_directory (char *dirname, struct file *f, int *N)
{
    DIR               *d;
    struct dirent     *e;
    int               i, N_max = *N;
    char              *dir, *p;

    *N = 0;
    dir = chunk_put (dirname);
    p = make_absolute (dirname);
    d = opendir (p);
    if (d == NULL) return 1;

    for (i=0; i<N_max; )
    {
        e = readdir (d);
        if (e == NULL) break;

        // skip some files
        if (ignore (e->d_name)) continue;

        f[i].dir  = dir;
        f[i].name = chunk_put (e->d_name);
        i++;
    }
    *N = i;
    closedir (d);

    fill_list (f, N);
    
    return 0;
}

/*--------------------------------------------------------------------*/

int fill_list (struct file *f, int *N)
{
    struct stat       ctm;
    char              buf1[4096], buf2[4096], *p;
    int               i, n;

    for (i=0; i<*N; i++)
    {
        f[i].type = TYPE_UNKNOWN;
        
        // obtain file information
        strcpy (buf1, f[i].dir);
        strcats (buf1, f[i].name);
        p = make_absolute (buf1);
        if (lstat (p, &ctm) < 0) continue;

        f[i].desc = NULL;  // for now
        
        // query item type
        if (S_ISREG(ctm.st_mode))
            f[i].type = TYPE_REGULAR;
        else if (S_ISDIR(ctm.st_mode))
            f[i].type = TYPE_DIRECTORY;
#ifndef __EMX__
        else if (S_ISLNK(ctm.st_mode))
            f[i].type = TYPE_SYMLINK;
#endif

#ifndef __EMX__
        // if it's symlink, then find out whether it's worth including
        // replace symlink's properties with destination's
        if (f[i].type == TYPE_SYMLINK)
           if (stat (p, &ctm) == 0)
           {
               if (S_ISDIR(ctm.st_mode) || S_ISREG(ctm.st_mode))
               {
                   n = readlink (p, buf2, sizeof (buf2));
                   buf2[n] = '\0';
                   f[i].desc = chunk_put (buf2);
               }
               else
               {
                   // ignore symlinks to something else
                   f[i].type = TYPE_UNKNOWN;
               }
           }
           else
           {
               // ignore hanging links
               fprintf (cgiOut, "!Hanging link : %s in %s\n", f[i].name, f[i].dir);
               f[i].type = TYPE_UNKNOWN;
           }
#endif
        
        // find out whether this dir has subdirs
        if (f[i].type == TYPE_DIRECTORY || f[i].type == TYPE_SYMLINK)
            f[i].hs = has_subdirs (buf1);
        else
            f[i].hs = 0;

        f[i].age  = ctm.st_mtime;
        f[i].size = ctm.st_size;
    }

    attach_descriptions (f, *N);
    qsort ((void *)f, *N, sizeof (struct file), filecompare1);

    // remove things not files or directories
    for (i=0; i<*N; i++)
        if (f[i].type == TYPE_UNKNOWN)
        {
            *N = i;
            break;
        }

    return 0;
}

/*--------------------------------------------------------------------*/

int resolve_link1 (struct file f)
{
    char  buffer[8192];
    char  *p, *p1, buf[8192];

    if (f.type != TYPE_SYMLINK) return -1;
    
    if (f.desc[0] == '/') return -1; // reject absolute links

    strcpy (buffer, f.dir);
    strcats (buffer, f.desc);
    
    while ((p = strstr (buffer, "/..")) != NULL)
    {
        strcpy (buf, p+3);
        *p = '\0';
        p1 = strrchr (buffer, '/');
        if (p1 == NULL)
        {
            strcpy (p, "/");
            strcat (p, buffer);
        }
        else
        {
            strcpy (p1, buf);
        }
    }
    
    return whereis (buffer);
}

/*--------------------------------------------------------------------
 returns index in fb.files array for specified path resolving symlinks
 along the way, or -1 if not found
 */

int whereis (char *path)
{
    char buf1[4096], buf2[4096], *p1;
    int  i, n;

    // sanity check
    if (path == NULL) return -1;  
    
    // remove starting '/'
    if (path[0] == '/')
        strcpy (buf1, path+1);
    else
        strcpy (buf1, path);

    // handle root directory separately
    if (buf1[0] == '\0') return 0;

    // break path into parts and verify them
    buf1[strlen(buf1)+1] = '\0';
    translate (buf1, '/', '\0');
    p1 = buf1;
    strcpy (buf2, "/");

    n = -1;
    while (*p1)
    {
        for (i=0; i<fb.N; i++)
            if (strcmp (fb.files[i].dir, buf2) == 0 &&
                strcmp (fb.files[i].name, p1) == 0)
                break;
        if (i == fb.N) return -1;
        switch (fb.files[i].type)
        {
        case TYPE_DIRECTORY:
        case TYPE_REGULAR:
            strcats (buf2, p1);
            break;
        case TYPE_SYMLINK:
            strcats (buf2, fb.files[i].desc);
            break;
        default:
            return -1;
        }
        p1 += strlen (p1) + 1;
        n = i;
    }

    if (n > 0 && fb.files[n].type == TYPE_SYMLINK)
        resolve_link (fb.files[n]);

    return n;
}

/*--------------------------------------------------------------------*/

int resolve_link (struct file f)
{
    char  buffer[8192];
    char  *p1, buf1[8192], buf2[8192];
    int   n, i;

    if (f.type != TYPE_SYMLINK) return -1;
    
    if (f.desc[0] == '/') return -1; // reject absolute links

    strcpy (buffer, f.dir);
    strcats (buffer, f.desc);
    squeeze_path (buffer);
    
    // remove starting '/'
    if (buffer[0] == '/')
        strcpy (buf1, buffer+1);
    else
        strcpy (buf1, buffer);

    // handle root directory separately
    if (buf1[0] == '\0') return 0;

    // break path into parts and verify them
    buf1[strlen(buf1)+1] = '\0';
    translate (buf1, '/', '\0');
    p1 = buf1;
    strcpy (buf2, "/");

    n = -1;
    while (*p1)
    {
        for (i=0; i<fb.N; i++)
            if (strcmp (fb.files[i].dir, buf2) == 0 &&
                strcmp (fb.files[i].name, p1) == 0)
                break;
        if (i == fb.N) return -1;
        switch (fb.files[i].type)
        {
        case TYPE_DIRECTORY:
        case TYPE_REGULAR:
            strcats (buf2, p1);
            break;
        case TYPE_SYMLINK:
            n = resolve_link (fb.files[i]);
            if (n < 0) return -1;
            strcpy (buf2, fb.files[n].dir);
            strcats (buf2, fb.files[n].name);
            break;
        default:
            return -1;
        }
        p1 += strlen (p1) + 1;
        n = i;
    }

    if (n > 0 && fb.files[n].type == TYPE_SYMLINK)
        return resolve_link (fb.files[n]);

    return n;
}

/* ----------------------------------------------------------- */

static int filecompare1 (__const__ void *elem1, __const__ void *elem2)
{
    struct file
        *f1 = (struct file *) elem1,
        *f2 = (struct file *) elem2;

    return (f2->type - f1->type);
}

